/* A* Algorithmus 
   die Menge der Knoten eines Graphen wird partitioniert in zwei Teile
   a) die schon besuchten (analysierten) Knoten und
   b) die noch zu analysierenden (die offenen) Knoten.

   Alle offenen Knoten erhalten eine Bewertung.
   Diese Bewertung setzt sich aus zwei Kostenfaktoren zusammen,
   den Wegekosten und einer Abschaetzung fuer die Guete des Knotens
   (fuer die Kosten vom Knoten zu einem Loesungsknoten).

   Die 2. Komponente ist eine Schaetzung.
   Sie wird durch eine heuristische Funktion berechnet.
   Das Wissen ueber das Problem wird in der heuristischen Funktion
   ausgenutzt.
   
   Datenstrukturen:
   die schon besuchten Knoten werden in einer Menge (Liste) gesammelt.
   Die offenen Knoten werden ebenfalls in einer Menge (Liste) gesammelt.
   Hier wird aber zusaetzlich der Pfad vom Knoten zum Startknoten und
   die Regelnamen und die Kosten fuer den Knoten gespeichert.

   Kosten sind als Paar g+h dargestellt, g sind die Wegekosten,
   h die Abschaetzung bis zum Ziel

   Ein offener Knoten hat die Gestalt p+c mit p als Pfadliste und  c als
   Kostenpaar, p hat die Form  n :-> r+p1 mit
   n als Zustans, r als erzeugnender Regel, p1 als Pfad vom
   Start zum Vorgaenger
*/

:-consult('trace.pl').

:- op(680,xfx,':->').

/*--- Kostenarithmetik ----------------------------------------------------------------*/

is_better(G0 + H0, G1 + H1) :-
	X0 is G0 + H0,
	X1 is G1 + H1,
	X0 =< X1.

min_cost(A0, A1, A0) :-
	A0 = _ + C0,
	A1 = _ + C1,
	is_better(C0, C1),
	!.
min_cost(_, A1, A1).

sel_best_node([A0], A0) :-
	!.

sel_best_node([A0 | R], A2) :-
	sel_best_node(R, A1),
	min_cost(A0, A1, A2).

/*--- generiere alle Nachfolger aus einem Knoten --------------------------------------*/

gen1(P, Succ) :-
        P = S :-> _ ,
        findall(S1 :-> R + P,
                rule(R, S, S1),
                Succ).

/*--- Mengenoperationen auf Zustandsmengen --------------------------------------------*/
  
union1(S, S1, S1) :-
	isin(S, S1),
	!.

union1(S, S1, [S | S1]).

                                                              /* minus(S1,S2,S1 - S2) */
minus([], _, []).

minus([H | R], S2, S3) :-
	isin(H, S2),
	!,
	minus(R, S2, S3).

minus([H | R], S2, [H | S3]) :-
	minus(R, S2, S3).
    
isin(S :-> _ , [S :-> _ | _]) :-
	!.
isin(P, [_ | T]) :-
	isin(P, T).

card([], 0).
card([_ | R], I) :-
	card(R, I1),
	I is I1 + 1.

/*--- Loesung suchen ------------------------------------------------------------------*/

test_solution(Open, _, Best, N, _, found(N, P, Open)) :-
	Best = P + _,
	P = S :-> _,
	solution(S),
	!.

test_solution(Open, Visited, Best, N, Max, Solution) :-
	Best = P + _,
	gen1(P, Succ),
	union1(P, Visited, Newvisited),
	minus(Succ, Newvisited, Newsucc),
	del_open_node(Open, Best, Open1),   
	add_open_nodes(Open1, Newsucc, Best, Newopen),
        nl,
	N1 is N + 1,
	solve(Newopen, Newvisited, N1, Max, Solution).

solve([], _, N, _, empty(N)) :-
	!.

solve(_, _, Max, Max, limit(Max)) :-
	!.
       
solve(Open, Visited, N, Max, Solution) :-
	sel_best_node(Open, Best),
	traceout((outattr_state(Best))
	),
	test_solution(Open, Visited, Best, N, Max, Solution).

solve(Initial, Max, Solution) :-
	solve([(Initial :-> nil) + (0+0)], [], 0, Max, Solution).

solve(Initial, Max) :-
	solve(Initial, Max, _Solution).

/*--- Mengenoperationen auf offenen Knoten --------------------------------------------*/

del_open_node([], _, []).

del_open_node([N | R], N, R) :-
	!.

del_open_node([H | R], N, [H | R1]) :-
	del_open_node(R, N, R1).
  
add_open_nodes(Open, [], _, Open).

add_open_nodes(Open, [S | R], Node, Newopen) :-
	add_open_nodes(Open, R ,Node, Newopen1),
	add_open_node(Newopen1, S ,Node, Newopen ).      

add_open_node(Open, S1, Best, Newopen) :-
	Best = S + (G0 + _),
	pathcost(S, S1, G1),
	G is G0 + G1,
	guess(S1, H),
        traceout((outattr_state(S1 + (G + H))) 
        ),
	ins_open_node(Open, S1 + (G + H), Newopen).

                                   /* bei gleichem Zustand und unterschiedlichen Wegen */
                                            /* wird der kostenguenstigere Weg gewaehlt */
ins_open_node([], N, [N]).
ins_open_node([N0 | R], N, [N1 | R]) :-
	N0 = (S :-> _) + _,
	N  = (S :-> _) + _,
	min_cost(N0, N, N1),
	!.
     
ins_open_node([N0 | R], N, [N0 | Newopen]) :-
	ins_open_node(R, N, Newopen).

/*--- Ausgabe -------------------------------------------------------------------------*/

outsolution(limit(N)) :-
        !,
        write('Keine Loesung in '), write(N), write(' Suchschritten gefunden'), nl.

outsolution(empty(N)) :-
        !,
        write('Nach '), write(N), write('Schritten Suchraum erschoepft'), nl.

outsolution(found(N, Solution, Open)) :-
	!,
        write('Nach '), write(N), write(' Schritten eine Loesung gefunden'), nl,
	write('Loesung'), nl,
	outpath(Solution), nl,
	traceout((card(Open,N1),
		write('noch offene Knoten: '),
		write(N1),nl)
	).

outopen_nodes([]).
outopen_nodes([N | R]) :-
	write('offen : '),
	outattr_state(N),
	outopen_nodes(R).
     
outattr_state((S :-> _) + C) :-
	outstate(S),
	writes(C).

outpath(S:->nil) :-
        outstate(S).

outpath(S:->R+P) :-
        outpath(P),
        outrule(R),
        outstate(S).

/*--- Standard Heuristiken ------------------------------------------------------------*/

:- assert(pathcost(_, _, 1)).
:- assert(guess(_, 0)).

setrule(Pred,Rule) :-
        retractall(Pred),
        assert(Rule).

kostenkonstant(C) :-
        setrule(pathcost(_,_,_),pathcost(_, _, C)).

keineheuristik :-
        setrule(guess(_,_), guess(_, 0)).

breitensuche :-
        kostenkonstant(1),
        keineheuristik.

tiefensuche :-
        kostenkonstant(-1),
        keineheuristik.

/*-------------------------------------------------------------------------------------*/

